"""MicroPython toy car controls."""
from array import array
import gc
from jq6500 import Player
from machine import ADC, I2C, Pin, SPI  # type: ignore
from math import cos, radians, sin
from micropython import const  # type: ignore
from ssd1309 import Display
from sx1509 import Expander, PinModes
from time import sleep_ms
from xglcd_font import XglcdFont


class LightMode:
    OFF = 0  # Lights off
    ON = 1  # Lights on
    BLINK = 2  # Lights blink


class Car(object):
    """Car controls."""

    # Constants
    CIRCLE_RADIUS = const(28)  # Circle encompassing wheel indicator
    TRIANGLE_BASE = const(20)  # Triangle wheel indicator base
    TRIANGLE_HEIGHT = const(27)  # Triangle wheel indicator height
    TRI_POSITIONS = const(17)  # Number of wheel indicator positions
    ANGLE_SPAN = 225 / (TRI_POSITIONS - 1)  # Wheel indicator angle span
    WHEEL_ADC = const(26)  # Steering wheel potentiometer ADC pin
    HYSTERESIS_THRESHOLD = const(1500)  # Wheel analog hysteresis threshold
    BLACK_MINI = const(10)  # Black miniature pushbutton (not on expander)
    GREEN_RING = const(11)  # Green metal ring pushbutton (not on expander)
    G1, G2, G3, G4 = const(19), const(18), const(17), const(16)  # Shifter pins
    ROCKER_HL1, ROCKER_HL2 = const(15), const(14)  # Headlights rockers
    ROCKER_STROBE1, ROCKER_STROBE2 = const(13), const(12)  # Strobe rockers
    HL1_LED, HL2_LED = const(6), const(7)  # Headlight LED expander GPIOs
    STROBE1_LED, STROBE2_LED = const(4), const(5)  # Strobe LED expander GPIOs
    GREEN_TRI_LED = const(15)  # Green triangular pushbutton LED expander GPIO
    LO_RIDER_LED = const(14)  # Lo Rider pushbutton LED expander GPIO
    CRUISIN_LED = const(13)  # Cruis'n pushbutton LED expander GPIO
    SKY_HI_LED = const(12)  # Sky Hi pushbutton LED expander GPIO
    GREEN_ILL_LED = const(6)  # Green illuminated pushbutton LED GPIO
    GREEN_METAL_LED = const(7)  # Green metal pushbutton ring LED GPIO
    CS = const(5)  # Display chip select pin
    DC = const(27)  # Display data/command mode pin
    RESET = const(28)  # Display reset pin
    KEY_ROWS = const(4)  # Number of rows in the keypad
    KEY_COLS = const(4)  # Number of columns in the keypad
    SLEEP_TIME = const(256)  # Time in ms for keypad low-current sleep mode
    SCAN_TIME = const(30)  # Scan time in ms devoted to each matrix row
    DEBOUNCE_TIME = const(15)  # Keypad debounce time in ms
    MAX_RELEASE = const(5)  # Number of cycles before a button can be released

    # Map of buttons to rows / columns
    KEY_MAP = [
        ['green_arcade', 'red_square', 'red_rs', 'red_strobe'],
        ['cruisn', 'green_tri', 'black_rs', 'hl_right'],
        ['sky_hi', 'green_metal', 'green_ill', 'hl_left'],
        ['lo_rider', 'yellow_arcade', 'red_round', 'green_strobe']
    ]

    # Dict to process sounds and LED's associated with button presses
    PROCESS_MAP = {
        'green_arcade': {'MP3': 'Hello Bradley', 'LED': 0},
        'cruisn': {'MP3': 'Tip Toes', 'LED': CRUISIN_LED},
        'sky_hi': {'MP3': 'Slide2', 'LED': SKY_HI_LED},
        'lo_rider': {'MP3': 'La Cucaracha', 'LED': LO_RIDER_LED},
        'red_square': {'MP3': 'Slide', 'LED': 0},
        'green_tri': {'MP3': 'Road Runner', 'LED': GREEN_TRI_LED},
        'green_metal': {'MP3': 'Scale', 'LED': 0},
        'yellow_arcade': {'MP3': 'Duck', 'LED': 0},
        'red_rs': {'MP3': 'Fanfare', 'LED': 0},
        'black_rs': {'MP3': 'Cuckoo', 'LED': 0},
        'green_ill': {'MP3': 'Engine', 'LED': GREEN_ILL_LED},
        'red_round': {'MP3': 'Pop', 'LED': 0},
        'red_strobe': {'MP3': 'Dog', 'LED': 0},
        'hl_right': {'MP3': 'Pig', 'LED': 0},
        'hl_left': {'MP3': 'Sheep', 'LED': 0},
        'green_strobe': {'MP3': 'Cat', 'LED': 0},
        'green_ring': {'MP3': 'Clown', 'LED': GREEN_METAL_LED},
        'black_mini': {'MP3': 'Boing', 'LED': 0}
    }

    # List of sounds (The list index + 1 corresponds to JQ6500 player index)
    SOUNDS = ["1", "2", "3", "4", "Boing", "Cat", "Clown", "Cuckoo", "Dog",
              "Duck", "Engine", "Fanfare", "Hello Bradley", "La Cucaracha",
              "Pig", "Pop", "Road Runner", "Scale", "Sheep", "Slide", "Slide2",
              "Tip Toes"]

    def __init__(self):
        """Constructor for car."""
        # Store triangle coordinates for each wheel position in dict
        self.triangle_coords = {}
        for wheel in range(self.TRI_POSITIONS):
            self.triangle_coords[self.TRI_POSITIONS - wheel - 1] = (
                self.calc_triangle_vertices(wheel))
        self.prev_wheel = 0  # Previous steering wheel analog input
        self.adc = ADC(26)  # Create an ADC for steering wheel
        self.unispace = XglcdFont('fonts/Unispace12x24.c', 12, 24)
        # Initialize SSD1309 display on SPI bus 0
        self.spi = SPI(0, baudrate=10000000, sck=Pin(2), mosi=Pin(3))
        self.display = Display(self.spi, dc=Pin(self.DC), cs=Pin(self.CS),
                               rst=Pin(self.RESET))

        # Initialize MP3 sound module on serial port 1
        self.player = Player(port=1, tx=8, rx=9)

        # Initialize gear GPIO pins as inputs with pull-ups.
        self.gear1 = Pin(self.G1, Pin.IN, Pin.PULL_UP)
        self.gear2 = Pin(self.G2, Pin.IN, Pin.PULL_UP)
        self.gear3 = Pin(self.G3, Pin.IN, Pin.PULL_UP)
        self.gear4 = Pin(self.G4, Pin.IN, Pin.PULL_UP)
        self.prev_gear = ""  # Last gear state

        # Initialize headlights and strobe rocker switches
        self.rocker_hl1 = Pin(self.ROCKER_HL1, Pin.IN, Pin.PULL_UP)
        self.rocker_hl2 = Pin(self.ROCKER_HL2, Pin.IN, Pin.PULL_UP)
        self.rocker_strobe1 = Pin(self.ROCKER_STROBE1, Pin.IN, Pin.PULL_UP)
        self.rocker_strobe2 = Pin(self.ROCKER_STROBE2, Pin.IN, Pin.PULL_UP)
        self.prev_headlights = LightMode.OFF  # Last headlights state
        self.prev_strobes = LightMode.OFF  # Last strobe state

        # Initialize SX1509 I/O expander on I2C bus 1
        self.i2c = I2C(0, freq=400000, scl=Pin(1), sda=Pin(0))
        self.expander = Expander(self.i2c)

        # Intialize expander LED's
        self.expander.pin_mode(self.HL1_LED, PinModes.OUTPUT)
        self.expander.pin_mode(self.HL2_LED, PinModes.OUTPUT)
        self.expander.pin_mode(self.STROBE1_LED, PinModes.OUTPUT)
        self.expander.pin_mode(self.STROBE2_LED, PinModes.OUTPUT)

        # Initialize additional LED's
        self.initialize_leds()

        # Set up pushbuttons not on expander
        self.black_mini = Pin(self.BLACK_MINI, Pin.IN, Pin.PULL_UP)
        self.green_ring = Pin(self.GREEN_RING, Pin.IN, Pin.PULL_UP)
        self.black_mini_pressed = False
        self.green_ring_pressed = False

        self.prev_key_data = 0  # Track previous keypad data
        self.release_count = 0  # Track loop cycles that button is released

        # Set up the keypad engine
        self.expander.keypad(self.KEY_ROWS, self.KEY_COLS, self.SLEEP_TIME,
                             self.SCAN_TIME, self.DEBOUNCE_TIME)

    def calc_triangle_vertices(self, wheel):
        """Calculate triangle vertices for a given wheel position.

        Args:
            wheel (int): The pin to configure for blinking
        Notes:
            Values are precalculated to improve performance.
        """
        angle = radians(112.5 - wheel * self.ANGLE_SPAN)
        x_apex = self.TRIANGLE_HEIGHT * sin(angle)
        y_apex = -self.TRIANGLE_HEIGHT * cos(angle)
        x_base_left = (-self.TRIANGLE_BASE / 2 * cos(angle) -
                       self.CIRCLE_RADIUS * sin(angle))
        y_base_left = (-self.TRIANGLE_BASE / 2 * sin(angle) +
                       self.CIRCLE_RADIUS * cos(angle))
        x_base_right = (self.TRIANGLE_BASE / 2 * cos(angle) -
                        self.CIRCLE_RADIUS * sin(angle))
        y_base_right = (self.TRIANGLE_BASE / 2 * sin(angle) +
                        self.CIRCLE_RADIUS * cos(angle))
        return array('h', [int(x_apex), int(y_apex),
                     int(x_base_left), int(y_base_left),
                     int(x_base_right), int(y_base_right)])

    def hello_bradley(self):
        """Display Hello Bradley image."""
        gc.collect()
        self.display.draw_bitmap('images/hello128X57.mono', 0, 7, 128, 57)
        self.display.present()

    def initialize_leds(self):
        """Initialize LED's in process mapping"""
        for pushbutton, config in self.PROCESS_MAP.items():
            led_pin = config['LED']
            if 1 <= led_pin <= 7:
                # Pico GPIO pin
                Pin(led_pin, Pin.OUT).value(0)
            elif 12 <= led_pin <= 15:
                # Expander GPIO pin
                self.expander.pin_mode(led_pin, PinModes.OUTPUT)

    def main(self):
        """Main program loop."""
        self.hello_bradley()
        sleep_ms(500)
        self.play_sound("Hello Bradley")
        sleep_ms(3000)
        self.display.clear_buffers()
        while True:
            # Steering Wheel Display
            self.display.fill_rectangle(0, 0, 64, 64, invert=True)
            self.display.draw_circle(31, 36, 27)
            wheel = self.adc.read_u16()  # Read 16 bit ADC value for wheel
            # Apply hysteresis
            if abs(wheel - self.prev_wheel) > self.HYSTERESIS_THRESHOLD:
                # The pot range is approximately 2000 to 48000
                wheel_scaled = wheel // 2706  # Scale ADC output 0-16
                wheel_scaled = min(wheel_scaled, 16)  # Catch errant readings
                self.prev_wheel = wheel  # Store last clean reading
            coords = self.triangle_coords[wheel_scaled]
            self.display.draw_polygon_coords(31, 36, coords, fill=True)

            # Gear shift display
            gear = "N"
            if self.gear1.value() == 0:
                gear = "1"
            elif self.gear2.value() == 0:
                gear = "2"
            elif self.gear3.value() == 0:
                gear = "3"
            elif self.gear4.value() == 0:
                gear = "4"
            self.display.draw_letter(95, 26, gear, self.unispace)
            if gear != self.prev_gear:
                if gear != "N" and self.prev_gear != "":
                    self.play_sound(gear)
                self.prev_gear = gear

            self.display.present()

            # Headlights and strobe lights rockers
            headlights = ((not self.rocker_hl1.value()) +
                          (not self.rocker_hl2.value()))
            strobes = ((not self.rocker_strobe1.value()) +
                       (not self.rocker_strobe2.value()))
            if strobes != self.prev_strobes:
                self.prev_strobes = strobes
                if strobes == LightMode.ON:
                    self.expander.setup_blink(self.STROBE1_LED, 0, 0, 0, 0,
                                              0, 0)
                    self.expander.setup_blink(self.STROBE2_LED, 0, 0, 0, 0,
                                              0, 0)
                elif strobes == LightMode.BLINK:
                    self.expander.breathe(self.STROBE1_LED, 500, 600, 600, 250)
                    self.expander.breathe(self.STROBE2_LED, 500, 600, 600, 250)
                else:
                    self.expander.setup_blink(self.STROBE1_LED, 0, 0, 255, 0,
                                              0, 0)
                    self.expander.setup_blink(self.STROBE2_LED, 0, 0, 255, 0,
                                              0, 0)
            if headlights != self.prev_headlights:
                self.prev_headlights = headlights
                if headlights == LightMode.ON:
                    self.expander.setup_blink(self.HL1_LED, 0, 0, 0, 0, 0, 0)
                    self.expander.setup_blink(self.HL2_LED, 0, 0, 0, 0, 0, 0)
                elif headlights == LightMode.BLINK:
                    self.expander.blink(self.HL1_LED, 500, 1000)
                    self.expander.blink(self.HL2_LED, 500, 1000)
                else:
                    self.expander.setup_blink(self.HL1_LED, 0, 0, 255, 0, 0, 0)
                    self.expander.setup_blink(self.HL2_LED, 0, 0, 255, 0, 0, 0)

            # Scan I/O expander keypad
            key_data = self.expander.read_keypad()
            if key_data != self.prev_key_data and key_data != 0:
                self.prev_key_data = key_data
                self.release_count = 1
                # Find the active row and columns
                row = self.expander.get_row(key_data)
                col = self.expander.get_col(key_data)
                self.process_input(self.KEY_MAP[row][col])
            elif self.prev_key_data != 0:
                if key_data == 0:  # Increment release count if no button press
                    self.release_count += 1
                else:  # Reset release count if button still pressed
                    self.release_count = 1
                if self.release_count > self.MAX_RELEASE:  # Release button
                    self.prev_key_data = 0

            # Poll additional 2 pushbuttons not on expander keypad
            button_state = self.read_debounced_pin(self.black_mini)
            if button_state == 1 and self.black_mini_pressed:
                self.black_mini_pressed = False  # Release
            elif button_state == 0 and not self.black_mini_pressed:
                self.process_input("black_mini")  # Pressed
                self.black_mini_pressed = True

            button_state = self.read_debounced_pin(self.green_ring)
            if button_state == 1 and self.green_ring_pressed:
                self.green_ring_pressed = False  # Release
            elif button_state == 0 and not self.green_ring_pressed:
                self.process_input("green_ring")  # Pressed
                self.green_ring_pressed = True

            sleep_ms(10)

    def play_sound(self, title):
        """Plays a sound effect on the MP3 module.

        Args:
            title (string): The sound effect title to play
        """
        try:
            index = self.SOUNDS.index(title) + 1
            self.player.play_by_index(index)
        except Exception as e:
            print("Error:", e)

    def process_input(self, button):
        """Handles sound effects and any LED effects for given button press.

        Args:
            button (string): Name of button or switch to process.
        """
        process = self.PROCESS_MAP[button]
        effect = process['MP3']
        self.play_sound(effect)
        led = process['LED']
        if led:
            self.set_led_state(led)
        # Handle Hello Bradley
        if effect == "Hello Bradley":
            self.hello_bradley()
            sleep_ms(3000)
            self.display.clear_buffers()

    def read_debounced_pin(self, pin, debounce_time=50):
        """Reads the GPIO pin state with debouncing.

        Args:
            pin: The pin to read
            debounce_time: Debounce time in milliseconds
        Returns: The debounced state of the pin
        """
        first_read = pin.value()
        sleep_ms(debounce_time)
        second_read = pin.value()
        return second_read if first_read == second_read else None

    def set_led_state(self, pin, off=False):
        """Either toggles or modifies LED state.

        Args:
            pin (int): The LED GPIO pin
            off (Optional bool): If True the LED will be turned off
        """
        if pin < 10:
            # LED connected to Pico GPIO
            led = Pin(pin, Pin.OUT)
            if off:
                led.value(0)
            else:
                led.value(not led.value())
        else:
            # LED connected to expander GPIO
            roff = self.expander.read_byte(self.expander.REG_OFF[pin])
            rion = self.expander.read_byte(self.expander.REG_I_ON[pin])
            if roff > 0 or off:
                # Currently breathing switch to off
                self.expander.setup_blink(pin, 0, 0, 255, 0, 0, 0)
            elif rion > 0:
                # Currently off, switch to on
                self.expander.setup_blink(pin, 0, 0, 0, 0, 0, 0)
            else:
                # Currently on, switch to breath
                self.expander.breathe(pin, 500, 600, 600, 250)


if __name__ == "__main__":
    print('starting program...')
    car = Car()
    try:
        car.main()
    except KeyboardInterrupt:
        print("Ctrl-C pressed.")
    finally:
        car.display.cleanup()
        car.expander.reset()
        car.set_led_state(car.GREEN_METAL_LED, off=True)
        car.set_led_state(car.GREEN_ILL_LED, off=True)
        print('Program terminated.')
